/**************************************************************************************

	FireMaster :  Firefox Master Password Recovery Tool
	Copyright (C) 2006  Nagareshwar Y Talekar ( tnagareshwar@gmail.com )

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

**************************************************************************************/



///////////////////////////////////////////////////////////////////////////
//
//
//    0x0x0x0        Firef0x Master Password Recovery Tool       0x0x0x0
//
//
//
//    Author  :  Nagareshwar Y Talekar 
//               tnagareshwar@gmail.com
//
//
//    Date    :  25th Dec 2005.
//
//	  Version :  1.0  ( 1st  Jan 2006   )
//               1.5  ( 14th Jan 2006   )
//               2.0  ( 25th Feb 2006   ) 
//               2.1  ( 18th Jun 2006   )
//				 2.2  ( 8th  July 2006  )
//
//    
//    Info    :  This program performs a dictionary/hybrid/bruteforce crack to get
//			 	 firefox's master password. One can copy key3.db from the
//				 firefox's profile directory and perform offline crack as well.
//
//				 Currently it supports Firefox 1.5 version only. It may or may not 
//				 work with higher version.
//    
//	  Usage   :  FireMaster <options> <firefox_profile_dir>
//
//    History :  25th Dec 2005   First development relese completed
//
//				 28th Dec 2005   Key3.db is cracked and password check algorithm is simplified.
//								 Passwords crack count per second is doubled. As a result
//								 total crack time is halved.
//
//				 29th Dec 2005   Optimized the softoken library by converting dynamic 
//								 memory allocations to static. Crack time reduced considerably.
//				
//				 1st Jan 2006    Further optimized the softoken and freebl library. Overall
//								 crack time reduced further.
//
//				 9th Jan 2006    Linked with release version of libraries. It doubled the speed 
//								 and reduced the time to half. Cracked the key3.db further which
//								 solved the bug in handling key3.db file.
//
//
//				 13th Jan 2006   Displaying the various results and also brute force time display
//								 corrected. Now user will get almost right time for brute force display
//								 based on their machine speed dynamically.
//								
//               15th Jan 2006   Major breakthrough has taken place. All library dependency has been removed
//							     Only required files are brought into main project. Optmizations and speed
//                               improvements....
//
//				 28th Jan 2006   Removed SHA1_Update recalculation and further optimized on that line
//		                            						 
//               31st Jan 2006   Lot of optimizations taken place . The speed on 2 setup is as follows
//								 Normal machine (1.6Ghz, 256Mb RAM ) : 50s 350ms : 48000
//                               High End m/c   (3GHz, Dual Core, 2GB RAM) : 23s 450ms : 104000
//
//               
//               23rd Feb 2006   Bug reported by T Barton <tbarton@lbcprof.com>.
//                               For perticular key3.db , initialization failed.
//								 Crack Speed Displayed at regular intervals.
//												 
//				 25th Feb 2006   Firemaster version 2.0 released.
//								 First public release of source code.
//				        
//				 27th Feb 2006   Removed minor defect in displaying dictCrack and hybrid crack
//								 Now it periodically displays statistics while operation is in progress.
//					
//               18th Jun 2006   Test successfully with latest firefox version 1.5.0.4.
//                               Some minor name changes here and there.
//			
//               8th July 2006   Source code is released under GPL. 
//
//
//   TODO :
//		  *) Speed improvements....High priority
//		  *) Add minimum length feature for bruteforce
//        *) use memory mapping for dictionary attacks
//		  *) Load the dictionary first completely ..then start cracking
//        *) Design GUI version  
//		  *) Port it to LINUX platform
//
//
//////////////////////////////////////////////////////////////////////////////////////

// Current speed 

// Normal    : 31st Jan 2006  : 50s 350ms : 48000
// High End  : 31st Jan 2006  : 23s 450ms : 104000




#include "KeyDBCracker.h"


// Global data...
KeyCrackData keyCrackData;
NSSPKCS5PBEParameter *paramPKCS5 = NULL;
//SECItem encStringItem;
SHA1Context pctx;
unsigned char encString[128];

	

// Parameters for brute force cracker
char bruteCharSet[256];
//char brutePasswd[100];
int brutePosCount;
int bruteCharCount;
double bruteCount = 0;
double totalBruteCount;



// Parameters for hybrid cracker
char hybridCharSet[256];
char hybridPasswd[100];
char hybridFinalPasswd[500];
int  hybridPosCount = 4;
int  hybridCharCount = 4;
int  hybridCount;
int  isSuffix = 1;

// Parameters for dict cracker
char *dictFile = NULL;
char dictPasswd[256];
char fileBuffer[51200];
int fileBufferSize = 51200;
double dictCount = 0;


char finalText[100];
int dictCrack=0;
int hybridCrack=0;
int bruteCrack=0;
int isQuiet = 0;


char *firefoxProfileDir;
int initTime, exitTime;
struct tm itime, *etime;



//
//  Checks if the specified password is correct with pw-check string
//
static int CheckMasterPassword(char *password)
{
unsigned char passwordHash[SHA1_LENGTH+1];

	SHA1Context ctx; 

	// Copy already calculated partial hash data..
	memcpy(&ctx, &pctx, sizeof(SHA1Context) );
	SHA1_Update(&ctx, (unsigned char *)password, strlen(password));
    SHA1_End(&ctx, passwordHash);  

	return nsspkcs5_CipherData(paramPKCS5, passwordHash, encString);  //&encStringItem );
}


//
//  Dictionary Cracker
//
//  if isHybrid flag is set then it attempts hybrid crack on each 
//  of the dictionary password
//
void DictCrack(char *dictFile,int isHybrid)
{
	
	FILE *f;
	int index, fileOffset=0;
	int i,j,readCount;
	int isLastBlock;

	if( isHybrid )
		printf("\n\n Performing hybrid crack %s.....please wait \n", isQuiet == 1?"(Quiet Mode)":"" );
	else
		printf("\n\n Performing dictionary crack %s....please wait \n", isQuiet == 1?"(Quiet Mode)":"" );

	f=fopen(dictFile,"rb"); //open word list

    if(!f)
	{
		printf("\nError opening dictionary file %s\n",dictFile);
		return;
	}

	fileOffset = 0;

    do
	{   
		// read bulk data from file....
		readCount = fread(fileBuffer, 1,fileBufferSize, f);

		if( readCount == 0 )
			break;
		
		// If we have read less chars..then this is the last block...
		if( readCount < fileBufferSize )
			isLastBlock = 1;
		else
			isLastBlock = 0;
		
		index = 0;
		
		// check if the begining contains 10,13 chars..if so just skip them...
		for(index=0; index < readCount && (fileBuffer[index]==13 || fileBuffer[index]==10) ; index++);

		do  
		{
		
			// Go through the file buffer..extracting each password....
			dictPasswd[0]=0;
			
			for(i=index,j=0; i < readCount && fileBuffer[i] != 13 ; i++,j++)
				dictPasswd[j]=fileBuffer[i];

			dictPasswd[j]=0;

			// check if reading finished before '13' i.e we hit the wall
			if( i >= readCount && !( isLastBlock && dictPasswd[0]!=0) )
			{
				
				if(fileBuffer[i] != 13 )
				{
					fileOffset += index;
					fseek(f, index-readCount,SEEK_CUR);  // negative makes it to move backward
				}
				else
				{
					fileOffset +=readCount;
				}
				
				break;
			}
			
			index += strlen(dictPasswd) + 2;
			
			dictCount++;

			if( !isQuiet )
				printf("\n [%s] %.0lf => Trying password : %s ", isHybrid?"HybridCrack":"DictCrack", dictCount, dictPasswd);
			else if( ((int)dictCount % 1000) == 0 )
				 printf("\n [%s] %.0lf => Trying password : %s ", isHybrid?"HybridCrack":"DictCrack", dictCount, dictPasswd);
				 
			
			if( CheckMasterPassword(dictPasswd) )
			{
				printf("\n\n ********  CR@CKED ********* "); 
				printf("\n The Firefox Master password is: %s",dictPasswd);
				printf("\n ***************************\n\n");

				fclose(f);
				FireMasterExit();
			}

			// if isHybrid flag set..then attempt hybrid crack ...
			if( isHybrid )
			{
				HybridCrack(0);
			}
		
		}
		while(1);

    }
	while(1); 
    
	fclose(f);
}



//
//  Hybrid Cracker
//
void HybridCrack(int index )
{

	if( index >= hybridPosCount )
	{
		return;
	}

	for(int i=0; i< hybridCharCount; i++ )
	{
		hybridPasswd[index] = hybridCharSet[i];
		hybridPasswd[index+1] = 0;
		hybridCount++;
		
		if( isSuffix == 1)  // attach at the end....
		{
			strcpy(hybridFinalPasswd, dictPasswd );
			strcat(hybridFinalPasswd, hybridPasswd);
		}
		else    // attach at the begining
		{
			strcpy(hybridFinalPasswd, hybridPasswd);
			strcat(hybridFinalPasswd, dictPasswd );
		}
	
		if( !isQuiet )
			printf("\n [hybrid] %d => Trying %s ", hybridCount, hybridFinalPasswd);
		
		if( CheckMasterPassword(hybridFinalPasswd) )
		{
			printf("\n\n ********  CR@CKED ********* "); 
			printf("\n The Firefox Master password is: %s",hybridFinalPasswd);
			printf("\n ***************************\n\n");

			FireMasterExit();
		}


		HybridCrack(index+1);
	}

}





//
//  Brute Force Cracker
//
void BruteCrack(const char *bruteCharSet, char *brutePasswd, const int index, int next )
{
int i;
int totalTime;

	if( index >= brutePosCount )
		return;
	
	for(i=0; i< bruteCharCount; i++ )
	{
		//brutePasswd[index] = bruteCharSet[i];
		brutePasswd[index] = bruteCharSet[next%bruteCharCount];
		brutePasswd[index+1] = 0;
		bruteCount++;
		next++;
		
		if( !isQuiet )
			printf("\n%.0lf = Attempting %s ", bruteCount, brutePasswd);
		
		if( ((int)bruteCount % 1000000) == 0 )
		{
			totalTime = GetTickCount() - initTime;

			printf("\n\n Completed password count = %.0lf , still remaining = %.0lf ", bruteCount, totalBruteCount - bruteCount);
			int bruteTime = ((int)totalBruteCount - (int)bruteCount) / MAX_CRACKS_PER_SECOND;
			printf("\n Remaining Time = %.2dd %.2dh %.2dm %.2ds ", (bruteTime/86400) , (bruteTime/3600) % 24, (bruteTime/60)%60, bruteTime % 60);
			printf("\n Brutecrack speed = %d cracks/sec",(int) ( bruteCount / (totalTime/1000) ) );
		}
		

		if( CheckMasterPassword(brutePasswd) )
		{
			printf("\n\n ********  CR@CKED ********* "); 
			printf("\n The Firefox Master password is: %s",brutePasswd);
			printf("\n ***************************\n\n");

			FireMasterExit();
		}
	
		BruteCrack(bruteCharSet, brutePasswd,index+1, next);
	}

}


//
//  
//
int FireMasterInit(char *dirProfile)
{
    SECItem saltItem;

	if( CrackKeyData(dirProfile, keyCrackData) == FALSE)
	{
		exit(0);	
	}

	// Initialize the pkcs5 structure...
	saltItem.type = (SECItemType) 0;
	saltItem.len  = keyCrackData.saltLen;
	saltItem.data = keyCrackData.salt;
	paramPKCS5 = nsspkcs5_NewParam(NULL, &saltItem, 1);
	
	if( paramPKCS5 == NULL)
	{
		printf("\n Failed to initialize NSSPKCS5 structure");
		exit(0);
    }

	// Current algorithm is 
	// SEC_OID_PKCS12_PBE_WITH_SHA1_AND_TRIPLE_DES_CBC

	// Setup the encrypted password-check string
    memcpy(encString, keyCrackData.encData, keyCrackData.encDataLen );
	
	//printf("\n checking if master password is set ..!");
	// check if master password is set ...!
	if( CheckMasterPassword("") == TRUE )
	{
		printf("\n Master password is not set ...exiting FireMaster \n\n");
		exit(0);
	}
	
	// Calculate partial sha1 data for password hashing...
    SHA1_Begin(&pctx);
	SHA1_Update(&pctx, keyCrackData.globalSalt, keyCrackData.globalSaltLen);
	
	initTime = GetTickCount();

	time_t longtime;
	time(&longtime);
	struct tm *t=localtime(&longtime);
	memcpy(&itime,t,sizeof(struct tm));

	//printf("\n FireMaster Init time : %.2d-%.2d-%.2d %.2d:%.2d:%.2d", itime->tm_mday, itime->tm_mon+1, itime->tm_year+1900, itime->tm_hour, itime->tm_min, itime->tm_sec);

	
	return TRUE;
}





void FireMasterExit()
{
int totalTime;
int totalms;
int msfactor;

	exitTime = GetTickCount();
	totalTime = exitTime - initTime;

	totalms = totalTime % 1000;
	msfactor = totalms > 500?1:0;

	// Get complete date and time....
	time_t longtime;
	time(&longtime);
	etime=localtime(&longtime);

	printf("\n\n FireMaster recovery operation statistics ");
	printf("\n ==========================================");
	if( bruteCrack )
	printf("\n Bruteforce crack speed : %d cracks/sec",(int) ( bruteCount / (totalTime/1000 + msfactor) ) );
	printf("\n FireMaster Init time   : %.2d-%.2d-%.2d %.2d:%.2d:%.2d", itime.tm_mday, itime.tm_mon+1, itime.tm_year+1900, itime.tm_hour, itime.tm_min, itime.tm_sec);
	printf("\n FireMaster Exit time   : %.2d-%.2d-%.2d %.2d:%.2d:%.2d", etime->tm_mday, etime->tm_mon+1, etime->tm_year+1900, etime->tm_hour, etime->tm_min, etime->tm_sec);
	printf("\n Total crack time       : %.2dh %.2dm %.2ds %.2dms \n\n",totalTime/3600000, (totalTime/60000) % 60, (totalTime / 1000) % 60, totalTime % 1000);
	
	exit(0);
}




void PrintUsage(char *progName)
{
	printf("\n\n Firef0x Master Password Recovery Version %s \n\t by Nagareshwar Y Talekar ", FIREMASTER_VERSION);

	printf("\n\n Usage : \n\n Firemaster [-q] ");
	printf("\n            [ [-d -f <file>] | ");
	printf("\n            [-h -f <file> [-c <count>] [-l \"charlist\"] [ -s | -p ] ] ] ");
	printf("\n            [-b -n <count> [-a \"charlist\"] ] ");
	printf("\n            \"<firefox_profile_dir>\" ");

	printf("\n\n Options : ");
	printf("\n -d        [dict] Perform dictionary crack  ");
	printf("\n -h        [hybrid] Perform hybrid crack . In addition to dictionary passwords ");
	printf("\n           it can find passwords like pass123, 12pass etc");
	printf("\n -f        [dict/hybrid] Dictionary file with words on each line");
	printf("\n -c        [hybrid] Specify the count of positions ");
	printf("\n -s        [hybrid] Suffix the generated chars to the wordlist (pass123)");
	printf("\n -p        [hybrid] Prefix the generated chars to the wordlist (123pass)");
	printf("\n -l        [hybrid] List of chars used for hybrid crack process");
	printf("\n -b        [bruteforce] Perform bruteforce crack  ");
	printf("\n -n        [bruteforce] No of positions for which to generate the passwd");
	printf("\n -a        [bruteforce] Array of chars used for bruteforce crack process");
	printf("\n -q        Quiet or Quick mode , don't print during crack operation");
	
	printf("\n\n Example ( possible combinations ) ");
	
	printf("\n\n // Only dictionary crack ");
	printf("\n %s -q -d -f dictfile.txt firefox_profile_dir", progName);
	
	printf("\n\n // Only hybrid crack");
	printf("\n %s -q -h -f dictfile.txt -c 3 -l \"123\" firefox_profile_dir", progName);

	printf("\n\n // Only bruteforce crack");
	printf("\n %s -q -b -n 6 -a \"abc123\" firefox_profile_dir ", progName);

	printf("\n\n // Dictionary crack followed by bruteforce crack");
	printf("\n %s -q -d -f dictfile.txt -b -n 6 firefox_profile_dir", progName);

	printf("\n\n // Hybrid crack followed by bruteforce crack ");
	printf("\n %s -q -h -f dictfile.txt -c 3 -l \"123\" -b -n 6 -a \"abc123\" firefox_profile_dir", progName);
	printf("\n\n");

	exit(0);

}




void ParseArg(int argc, char *argv[])
{

	if( argc < 3 )
		PrintUsage(argv[0]);

	// Last argument....
	argc -= 1;

	// Setup the default values....
	dictFile = NULL;
	dictCrack = 0;
	bruteCrack = 0;
	hybridCrack = 0;

	isSuffix = 1;
	hybridPosCount = 3;
	strcpy(hybridCharSet, "123");
	hybridCharCount	= strlen( hybridCharSet);

	brutePosCount = 6;
	strcpy(bruteCharSet, "abc123");
	bruteCharCount = strlen( bruteCharSet );

	isQuiet = 0;
	
	for(int i=1; i < argc; i++)
	{
		if( argv[i][1] == 'd' || argv[i][1] == 'D' )
			dictCrack = 1;

		if( argv[i][1] == 'h' || argv[i][1] == 'H' )
			hybridCrack = 1;
		
		if( argv[i][1] == 'b' || argv[i][1] == 'B' )
			bruteCrack = 1;
		
		if( argv[i][1] == 's' || argv[i][1] == 'S' )
			isSuffix = 1;

		if( argv[i][1] == 'p' || argv[i][1] == 'P' )
			isSuffix = 0;

		if( argv[i][1] == 'q' || argv[i][1] == 'Q' )
		{
			//printf("\n quiet mode enabled ***********");
			isQuiet = 1;
		}
		
		if( argv[i][1] == 'f' || argv[i][1] == 'F' )
		{
			// If file name not present ....printUsage and exit
			if( i+1 >= argc )
			{
				printf("\n\n Error : File name not specified with -f option");
				PrintUsage(argv[0]);
			}

			dictFile =argv[i+1];
			i++;
		}

		if( argv[i][1] == 'c' || argv[i][1] == 'C' )
		{
			// if count value not specified ...exit
			if( i+1 >= argc )
			{
				printf("\n\n Error : Value not specified with -c option");
				PrintUsage(argv[0]);
			}

			hybridPosCount = atoi(argv[i+1]);
			if( hybridPosCount == 0 )
			{
				printf("\n\n Error : Specify the proper numeric value with -c option");
				PrintUsage(argv[0]);
			
			}

			i++;
		}

		if( argv[i][1] == 'n' || argv[i][1] == 'N' )
		{
			// if count value not specified ...exit
			if( i+1 >= argc )
			{
				printf("\n\n Error : Value not specified with -n option");
				PrintUsage(argv[0]);
			}

			brutePosCount = atoi(argv[i+1]);

			if( brutePosCount == 0 )
			{
				printf("\n\n Error : Specify the proper numeric value with -n option");
				PrintUsage(argv[0]);
			
			}

			i++;
		}

		if( argv[i][1] == 'l' || argv[i][1] == 'L' )
		{
			// if count value not specified ...exit
			if( i+1 >= argc )
			{
				printf("\n\n Error : Char List is not specified with -l option");
				PrintUsage(argv[0]);
			}
			strcpy(hybridCharSet, argv[i+1]);
			hybridCharCount	= strlen( hybridCharSet);
			i++;
		}

		if( argv[i][1] == 'a' || argv[i][1] == 'A' )
		{
			// if count value not specified ...exit
			if( i+1 >= argc )
			{
				printf("\n\n Error : Char List is not specified with -a option");
				PrintUsage(argv[0]);
			}
			strcpy(bruteCharSet, argv[i+1]);
			bruteCharCount	= strlen( bruteCharSet);
			i++;
		}

	
	}

	if( !dictCrack && !bruteCrack && !hybridCrack )
	{
		printf("\n\n Error : No crack method specified ");
		PrintUsage(argv[0]);
	}

	// Both  dictCrack or hybridCrack method cannot be specified 
	// At a time only one can be specified
	if( dictCrack && hybridCrack )
	{
		printf("\n\n Error : Invalid crack method combination , specify either dictionary crack  or hybrid crack method");
		PrintUsage(argv[0]);
	}


	// check if file name is specified for dict/hybrid crack
	if( dictCrack || hybridCrack )
	{
		if( dictFile == NULL || dictFile[0]==0 )
		{
			printf("\n\n Error : Dictionary file is not specified ");
			PrintUsage(argv[0]);
		}	
	}

	firefoxProfileDir = argv[argc];
	DWORD attr = GetFileAttributes(firefoxProfileDir); 
	
	if( (attr == -1) || !(attr & FILE_ATTRIBUTE_DIRECTORY ) )
	{
		printf("\n Error : Check if the directory path is proper [%s] ", firefoxProfileDir);
		PrintUsage(argv[0]);
	}


	// Print out the parameters...
	printf("\n\n Firef0x Master Password Recovery Version %s \n\t by Nagareshwar Y Talekar \n\n ", FIREMASTER_VERSION);

	printf("\n Performing Firefox Master password recovery operation ......");
	
	printf("\n\n Firefox profile path : [%s]", firefoxProfileDir);
	if( dictCrack )
	{
		printf("\n\n Recovery Method :   Dictionary");
		printf("\n Dict  File      :   %s ", dictFile);
	}


	if( hybridCrack )
	{
		printf("\n\n Recovery Method :   Hybrid");
		printf("\n Dict  File      :   %s ", dictFile);
		printf("\n Position Count  :   %d ", hybridPosCount);
		printf("\n Char Count      :   %d ", hybridCharCount);
		printf("\n Hybrid Char Set :   [%s] ", hybridCharSet);
		printf("\n Attachment      :   %s ", isSuffix == 0?"Prefix" : "Suffix");
	}


	if( bruteCrack )
	{
		printf("\n\n Recovery Method :   Bruteforce ");
		printf("\n Position Count  :   %d ", brutePosCount);
		printf("\n Char Count      :   %d ", bruteCharCount);
		printf("\n Brute Char Set  :   [%s] ", bruteCharSet);
	}


	printf("\n\n Press any key to start the Master Password recovery operation...\n");

	getch();

}



//
//  Main entrance....
//
int main(int argc, char* argv[])
{
	

	ParseArg(argc,argv);
	
	FireMasterInit(firefoxProfileDir);

	//printf("\n Initialization over..");

	if( dictCrack || hybridCrack )  
		DictCrack(dictFile, hybridCrack);
	


	if( bruteCrack )
	{
		totalBruteCount = bruteCharCount * ( ( pow(bruteCharCount, brutePosCount) - 1 ) / (bruteCharCount -1 ) ) ;
		double bruteTime = totalBruteCount/MAX_CRACKS_PER_SECOND + 1;
		
		printf("\n\n Performing bruteforce crack %s...", isQuiet == 1?"(Quiet Mode)":"");
		printf("\n Total password count  = %.0lf ", totalBruteCount );
		printf("\n Total Bruteforce Time = %.2dd %.2dh %.2dm %.2ds (Assuming %d cracks per second) ", (int)(bruteTime/86400) , (int)(bruteTime/3600) % 24, (int)(bruteTime/60)%60, (int)bruteTime % 60, MAX_CRACKS_PER_SECOND);

		// force quiet mode
		if( totalBruteCount > 50000 )
			isQuiet = 1;

		if( !isQuiet )
		{
			printf("\n\n Press any key to continue the BruteCr@cK...\n");
			getch();
		}
		else
			printf("\n\n Bruteforce crack is in progress, please wait....");

		initTime = GetTickCount();
		char brutePassword[256]="";
		BruteCrack(bruteCharSet, brutePassword, 0,0);

	}

	
	printf("\n\n Password not found :( , try better options \n");

	FireMasterExit();

	return 0;
}
