/**************************************************************************************

	FireMaster :  Firefox Master Password Recovery Tool
	Copyright (C) 2006  Nagareshwar Y Talekar ( tnagareshwar@gmail.com )

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

**************************************************************************************/




#include "KeyDBCracker.h"

/*
  What may go wrong ????
=========================================
 *) Instead of NSS_init ..more optimized secoid_init is used.
	If something goes wrong..use NSS_INIT and comment out section below it.
 *) Softoken optimized library is used. If you suspect just specify softoken.lib
    instead of softoken_opt lib without any hesitation
 *) Some of the sizes in KeyCrackData structure are assumed based on observation.
    In future this sizes may vary depending upon algo used.
*/


FILE *keyFile = NULL;


BOOL OpenKeyDBFile(char *profilePath)
{
char *keydbFileName = (char*) malloc( strlen(profilePath) + strlen(KEYDB_FILENAME) + 5);

	if( keydbFileName == NULL )
	{
		printf("\n OpenKeyDBFile : Insufficient memory  ");
		return FALSE;	
	}

	strcpy(keydbFileName, profilePath);
	strcat(keydbFileName, "\\");
	strcat(keydbFileName, KEYDB_FILENAME);

	keyFile = fopen(keydbFileName, "rb");
	if( keyFile == NULL )
	{
		printf("\n Failed to open file %s ", keydbFileName);
		return FALSE;
	}

	return TRUE;
}



BOOL CrackKeyData(char *profilePath, KeyCrackData &keyCrackData)
{
unsigned char buffer[KEYDB_MAX_BLOCK_SIZE];
int index = 0;
int offset;
KeyCrackOffset offData;	
int KEYDB_MAGIC_OFFSET;
char off[2];

	if( OpenKeyDBFile( profilePath ) == FALSE)
		return FALSE;

	// First read the magic offset 
	// This offset is present at offset 0x000f from the file..
	fseek(keyFile, 0x000e, SEEK_SET);
	
	if( fread(&off, 2, 1, keyFile) < 1)
	{
		printf("\n Error in reading data from %s file",KEYDB_FILENAME);
		return FALSE;
	}

	KEYDB_MAGIC_OFFSET = (off[0] << 8) + off[1];
	
	// Go to magic offset and read the offset information block
	fseek(keyFile, KEYDB_MAGIC_OFFSET, SEEK_SET);

	if( fread(&offData, sizeof(KeyCrackOffset), 1, keyFile) < 1)
	{
		printf("\n Error in reading data from %s file",KEYDB_FILENAME);
		return FALSE;
	}

	// Read the global salt 
	offset =  (offData.glbSaltOff[0] << 8) + offData.glbSaltOff[1] + KEYDB_MAGIC_OFFSET;
	fseek(keyFile, offset, SEEK_SET);

	keyCrackData.globalSaltLen = 16;
	if( fread(&keyCrackData.globalSalt, keyCrackData.globalSaltLen, 1, keyFile) < 1)
	{
		printf("\n Error in reading data from %s file",KEYDB_FILENAME);
		return FALSE;
	}

	keyCrackData.globalSalt[keyCrackData.globalSaltLen]=0;
	
	// Now read the main block of information...
	offset =  (offData.mainOff[0] << 8) + offData.mainOff[1] + KEYDB_MAGIC_OFFSET;
	fseek(keyFile, offset, SEEK_SET);

	if( fread(buffer, KEYDB_MAX_BLOCK_SIZE, 1, keyFile) < 1)
	{
		printf("\n Error in reading data from %s file",KEYDB_FILENAME);
		return FALSE;
	}

	// Now extract the data and fill up the structure...
	
	// Read the version, salt length, nn length
	keyCrackData.version = buffer[index++];
	keyCrackData.saltLen = buffer[index++];
	keyCrackData.nnLen   = buffer[index++];

	// Copy the salt
	unsigned char *salt = (unsigned char*)malloc(keyCrackData.saltLen+1);
	memcpy(salt, &buffer[index], keyCrackData.saltLen);
	salt[keyCrackData.saltLen] =0;
	keyCrackData.salt = salt;

	index += keyCrackData.saltLen;

	// copy nick name
	unsigned char *nickName =  (unsigned char*) malloc(keyCrackData.nnLen+1);
	memcpy(nickName, &buffer[index], keyCrackData.nnLen);
	nickName[keyCrackData.nnLen] =0;
	keyCrackData.nickName = nickName;

	index +=keyCrackData.nnLen;

	// Copy OID stuff
	keyCrackData.oidLen = buffer[index++];
	unsigned char *oidData =  (unsigned char*) malloc(keyCrackData.oidLen+1);
	memcpy(oidData, &buffer[index], keyCrackData.oidLen);
	oidData[keyCrackData.oidLen] =0;
	keyCrackData.oidData = oidData;

	index +=keyCrackData.oidLen;

	// Copy encrypted string
	memcpy(keyCrackData.encData, &buffer[index], 16 );
	keyCrackData.encData[16] = 0;
	keyCrackData.encDataLen = 16;
	index += 16;

	// copy password check string .. currently not used
	unsigned char *pwCheckStr =  (unsigned char*) malloc( strlen(KEYDB_PW_CHECK_STR) + 1);
	memcpy(pwCheckStr, KEYDB_PW_CHECK_STR, strlen(KEYDB_PW_CHECK_STR));
	pwCheckStr[strlen(KEYDB_PW_CHECK_STR)] =0;
	keyCrackData.pwCheckStr = pwCheckStr;
	
	index += strlen(KEYDB_PW_CHECK_STR);


	//=== Just print here what he have got ....====
	/*
	int i;
	printf("\n Magic offset = %x ", KEYDB_MAGIC_OFFSET);
	printf("\n Version : %d ", keyCrackData.version);
	
	printf("\n Global salt is \n");
	for(i=0; i < 16 ; i++)
		printf("%.2x ",keyCrackData.globalSalt[i]);
	
	printf("\n Salt is \n");
	for(i=0; i < 16 ; i++)
		printf("%.2x ",keyCrackData.salt[i]);
	

	printf("\n Encrypted data  is \n");
	for(i=0; i < 16 ; i++)
		printf("%.2x ",keyCrackData.encData[i]);

	printf("\n Algorithm is : \n ");
	for(i=0; i < keyCrackData.oidLen ; i++)
		printf("%.2x ",keyCrackData.oidData[i]);
	*/

	
	fclose(keyFile);

	return TRUE;
}

